// dear emacs, please treat this as -*- C++ -*-

#include <R.h>
#include <Rmath.h>
#include <math.h>

  // parameters
  // process model parameters w/ log or logit transformation
#define DELTA             (p[parindex[0]]) // natural death rate, fixed parameter
#define NETA              (p[parindex[1]]) // number of etas for the spline
#define LOGZETA           (&p[parindex[2]]) // seasonal eta vector for VZimmunomodulation
#define LOGITOMEGA        (p[parindex[3]]) // the relative infectiousness of herpes zoster versus varicella zoster, logit between 0-1
#define LOGALPHA          (p[parindex[4]]) // exponent in the force of infection
#define LOGBETA_SD        (p[parindex[5]]) // standard deviation of gamma distributed process noise
#define PHI               (p[parindex[6]]) // the rate at which an exposed person becomes infected with varicella zoster and manifests as chickenpox, fixed parameter
#define GAMMA             (p[parindex[7]]) // the recovery rate for varicella zoster, fixed parameter
#define IOTA              (p[parindex[8]]) // the average rate at which latent infections reactivate to manifest as herpes zoster, fixed parameter
  // measurement model parameters, independent reporting for varicella zoster and herpes zoster
#define LOGITRHOVZ        (p[parindex[9]]) // the mean report rate for varicella zoster
#define LOGTAUVZ          (p[parindex[10]]) // observation model dispersion parameter for varicella zoster
#define LOGTAUHZ          (p[parindex[11]]) // observation model dispersion parameter for herpes zoster
#define LOGM1             (p[parindex[12]]) // observation model dispersion parameter for varicella zoster
#define LOGM2             (p[parindex[13]]) // observation model dispersion parameter for herpes zoster
#define LOGZETAB          (&p[parindex[14]]) // seasonal eta vector for HZ immunomodulation
#define LOGETA            (p[parindex[15]]) // boosting parameter
#define LOGOMICRON        (p[parindex[16]]) // boosting parameter
#define LOGCHI            (p[parindex[17]]) // scales UV to VZ sigmoid
#define LOGUPSILON        (p[parindex[18]]) // scales UV to VZ sigmoid
#define LOGSIGMA          (p[parindex[19]]) // scales UV to VZ sigmoid
#define LOGA              (p[parindex[20]]) // scales UV to HZ sigmoid
#define LOGNU             (p[parindex[21]]) // scales UV to HZ sigmoid
#define LOGXI             (p[parindex[22]]) // scales UV to HZ sigmoid
#define LOGMULT           (p[parindex[23]]) // scales UV to HZ sigmoid
#define LOGEXPONY         (p[parindex[24]]) // scales UV to HZ sigmoid
  // state variables
  // probabilities realting to state transitions
#define pS                 (x[stateindex[0]]) // prob(remaining in Susceptible class), this is time-varying
#define pE                 (x[stateindex[1]]) // prob(remaining in Exposed class)
#define pI                 (x[stateindex[2]]) // prob(remaining in Infected varicella zoster or herpes zoster class)
#define pL1                (x[stateindex[3]]) // prob(remaining in the first latent class)
#define pL2                (x[stateindex[4]]) // prob(remaining in the second latent class)
  // other state variables
#define FOI                (x[stateindex[5]]) // force of infection
#define S                  (x[stateindex[6]]) // susceptible
#define E                  (x[stateindex[7]]) // exposed
#define IVZ                (x[stateindex[8]]) // infected, varicella zoster
#define L1                 (x[stateindex[9]]) // latent after varicella zoster infecton
#define IHZ                (x[stateindex[10]]) // infected/reactive herpes zoster
#define L2                 (x[stateindex[11]]) // latent after herpes zoster infection
#define IVZ_NEW            (x[stateindex[12]]) // new infections with varicella zoster
#define IHZ_NEW            (x[stateindex[13]]) // new infections with herpes zoster
#define report_rate_hz     (x[stateindex[14]]) // time varying reporting for herpes zoster
  // covariates
#define N           (covar[covindex[0]]) // population size
#define BABIES      (covar[covindex[1]]) // number of births
#define C1          (covar[covindex[2]]) // covariate 1  for the trend in herpes zoster reporing
#define C2          (covar[covindex[3]]) // covariate 2 for the trend in herpes zoster reporing
#define UV          (covar[covindex[4]]) // UV covariate
#define UV2         (covar[covindex[5]]) // UV covariate ALREADY lagged 2 months
#define TT          (covar[covindex[6]]) // Term-Time Forcing covariates
#define MODELNUMBER (covar[covindex[7]]) // model number
#define BOOSTING    (covar[covindex[8]]) // boosting on/off
#define SEAS        (&covar[covindex[9]]) // basis vector for seasonal VZ immunomodulation
#define SEASB       (&covar[covindex[10]]) // basis vector for seasonal HZ immunomodulation
  // observations
#define VARICELLA_ZOSTER    (y[obsindex[0]]) // observed cases of varicella zoster (i.e., chicken pox)
#define HERPES_ZOSTER       (y[obsindex[1]]) // observed cases of herpes zoster (i.e., shingles)

// logit function for transformations
// static double logit (double p) {
//   return log(p/(1-p));
// }

// expit function for transformations
static double expit (double x) {
  return 1.0/(1.0+exp(-x));
}

// measurement density
void zoster_meas_dens (double *lik, double *y, double *x, double *p, int give_log,
      int *obsindex, int *stateindex, int *parindex, int *covindex,
      int ncovar, double *covar, double t) {

  //define doubles
  double report_rate_vz;
  double tau_vz;
  double tau_hz;
  double lik1;
  double lik2;
  double templik;
  double tol = 1.0e-18;

  // untransform reporting parameters
  report_rate_vz = expit(LOGITRHOVZ);
  tau_vz = exp(LOGTAUVZ);
  tau_hz = exp(LOGTAUHZ);

  // Varicella zoster likelihood
  if(VARICELLA_ZOSTER > 0.0){
    lik1 = pnorm(VARICELLA_ZOSTER+0.5,report_rate_vz*IVZ_NEW,tau_vz*IVZ_NEW,1,0) - pnorm(VARICELLA_ZOSTER-0.5,report_rate_vz*IVZ_NEW,tau_vz*IVZ_NEW,1,0); 
  } else{
    lik1 = pnorm(VARICELLA_ZOSTER+0.5,report_rate_vz*IVZ_NEW,tau_vz*IVZ_NEW,1,0);
  }

  // Herpes zoster likelihood
  if(HERPES_ZOSTER > 0.0){
    lik2 = pnorm(HERPES_ZOSTER+0.5,report_rate_hz*IHZ_NEW,tau_hz*IHZ_NEW,1,0) - pnorm(HERPES_ZOSTER-0.5,report_rate_hz*IHZ_NEW,tau_hz*IHZ_NEW,1,0); 
  } else{
    lik2 = pnorm(HERPES_ZOSTER+0.5,report_rate_hz*IHZ_NEW,tau_hz*IHZ_NEW,1,0);
  }

   // work on log-scale bc computer can hard time with super small numbers
   // calculate the likelihood of observing the cases in both data sets, call it templik
    templik = exp(log(lik1) + log(lik2));
    *lik = templik + tol;

  // take the log of the likelihood
  if (give_log) *lik = log(*lik);
  if (!isfinite(*lik)) Rprintf("zoster_meas_dens %lg %lg %lg\n",VARICELLA_ZOSTER,HERPES_ZOSTER,*lik);
}

// measurement model 
void zoster_meas_sim (double *y, double *x, double *p, 
           int *obsindex, int *stateindex, int *parindex, int *covindex,
           int ncovar, double *covar, double t) {

  //define doubles
  double report_rate_vz;
  double tau_vz;
  double tau_hz;

  // untransform reporting parameters
  report_rate_vz = expit(LOGITRHOVZ);
  tau_vz = exp(LOGTAUVZ);
  tau_hz = exp(LOGTAUHZ);

  VARICELLA_ZOSTER = rnorm(report_rate_vz*IVZ_NEW,tau_vz*IVZ_NEW);
  if (VARICELLA_ZOSTER >= 0.0) {
    VARICELLA_ZOSTER = nearbyint(VARICELLA_ZOSTER);
  } else {
    VARICELLA_ZOSTER = 0.0;
  }

  HERPES_ZOSTER = rnorm(report_rate_hz*IHZ_NEW,tau_hz*IHZ_NEW);
  if (HERPES_ZOSTER >= 0.0) {
    HERPES_ZOSTER = nearbyint(HERPES_ZOSTER);
  } else {
    HERPES_ZOSTER = 0.0;
  }
}

// process model
void zoster_proc_sim (double *x, const double *p, 
         const int *stateindex, const int *parindex, const int *covindex,
         int covdim, const double *covar, 
         double t, double dt)
{
 // define doubles 
 double zeta; 
 double omega;
 double alpha;
 double beta_sd;
 double epsilon;
 double scale;
 double m1;
 double m2;
 double zetaB;
 double eta;
 double omicron;
 double scalarvz; //still used to equal sigmoid
 double scalarhz; //still used to equal sigmoid
 double scalarhz2; //still used to equal sigmoid
 double BetaT;
 double kappa;
 double chi;
 double upsilon;
 double sigma;
 double a;
 double nu;
 double xi;
 double psi;
 double mult;
 double expony;
 double TT_rel;

  // define integers
 int neta = (int) NETA;
 int j;
 int i;
 // int fail;

 // construct the B-spline
 for (j = 0, zeta = 0; j < neta; j++)  //j++ means go one plus in the loop from 0->5
    zeta += LOGZETA[j]*SEAS[j]; //logeta vector with 6 coefficients for BSpline, seas -> 6 covariates, this is the BSpline basis
    zeta = exp(zeta);
 // construct the second B-spline
 for (i = 0, zetaB = 0; i < neta; i++)  //j++ means go one plus in the loop from 0->5
    zetaB += LOGZETAB[i]*SEASB[i]; //logeta vector with 6 coefficients for BSpline, seas -> 6 covariates, this is the BSpline basis
    zetaB = exp(zetaB);

// untransform parameters for use in the model
 omega = expit(LOGITOMEGA);
 alpha = expit(LOGALPHA);
 beta_sd = exp(LOGBETA_SD);
 m1 = exp(LOGM1);
 m2 = exp(LOGM2);
 eta = exp(LOGETA);
 omicron = exp(LOGOMICRON);
 chi = exp(LOGCHI);
 upsilon = exp(LOGUPSILON);
 sigma = exp(LOGSIGMA);
 a = exp(LOGA);
 nu = exp(LOGNU);
 xi = exp(LOGXI);
 mult = exp(LOGMULT);
 expony = exp(LOGEXPONY);

//estimate seasonal susceptibility or reactivation from UV
scalarvz = 1/(chi+exp((upsilon*UV)-sigma));
scalarhz =  1/(a+exp((nu*UV)-xi));
scalarhz2 = 1/(a+exp((nu*UV2)-xi));

TT_rel = mult*pow(TT,expony);

// calculate the gamma distributed process noise epsilon
  if (beta_sd > 0.0) {
    // calculte the scale parameter, which here is also the variance
    // in C, rgamma(shape,scale), if shape==(1/scale), then mean==shape*scale==1 & variance==shape*scale*scale=scale
     scale = pow(beta_sd,2);
     epsilon = rgamma(1/scale,scale); 
  } else {epsilon = 1.0;}

  // run some checks, if non-finte values return and do not continue
  if(
     isfinite(zeta)== FALSE ||
     isfinite(epsilon)== FALSE ||
     isfinite(omega)==FALSE ||
     isfinite(alpha)==FALSE ||
     isfinite(FOI)==FALSE ||
     isfinite(pS)==FALSE ||
     isfinite(pL1)==FALSE ||
     isfinite(S)==FALSE ||
     isfinite(IVZ)==FALSE ||
     isfinite(IHZ)==FALSE ||
     isfinite(zetaB)==FALSE ||
     isfinite(eta)==FALSE ||
     isfinite(omicron)==FALSE ||
     isfinite(scalarvz)==FALSE ||
     isfinite(scalarhz)==FALSE ||
     isfinite(scalarhz2)==FALSE ||
     isfinite(TT_rel)==FALSE
     )
    {
      Rprintf("non finite value in zoster_proc_sim\n");
      return;
    }

  // process model equations
    if(MODELNUMBER == 1){
      BetaT = zeta; kappa = zeta;
    }else if(MODELNUMBER == 2){
      BetaT = zeta; kappa = zetaB;
    }else if(MODELNUMBER == 3){
      BetaT = zeta; kappa = scalarhz;
    }else if(MODELNUMBER == 4){
      BetaT = scalarvz; kappa = scalarhz;
    }else if(MODELNUMBER == 5){
      BetaT = zeta; kappa = scalarhz2;
    }else if(MODELNUMBER == 6){
      BetaT = TT_rel; kappa = zetaB;
    }else //if(MODELNUMBER == 7)
    {
      BetaT = TT_rel; kappa = scalarhz2;
    //}else {Rprintf("You messed up your Model numbering\n");
  }

    // if then for antibody boosting on/off
       if(BOOSTING == 1){
         psi = 1/(1+exp(eta*IVZ-omicron));
       }else //(BOOSTING == 0)
       {
         psi = 1;
       // }else{Rprintf("You messed up your Boosting numbering\n");  
       }

// General form for FOI
  FOI = BetaT*pow(((IVZ+omega*IHZ)/N),alpha)*epsilon;         
  if(FOI < 0.0){FOI = 0.0;}
  // probabilites based on the per capita rates and the time step, dt in months
  // we intend to use a dt of 1/4 month, i.e., weekly time steps
  pS = exp(-dt*(FOI+DELTA));
  pE = exp(-dt*(PHI+DELTA));
  pI = exp(-dt*(GAMMA+DELTA));
  pL1 = exp(-dt*(IOTA*kappa*psi+DELTA));
  pL2 = exp(-dt*(DELTA));
  // epi classes
  S = dt*BABIES + S*pS;
  E = (S*(1-pS)*FOI/(DELTA+FOI)) + (pE*E);
  IVZ = (E*(1-pE)*PHI/(DELTA+PHI)) + (pI*IVZ);
  L1 = (IVZ*(1-pI)*GAMMA/(GAMMA+DELTA)) + (pL1*L1);
  IHZ = (L1*(1-pL1)*(IOTA*kappa*psi)/(IOTA*kappa*psi+DELTA)) + (pI*IHZ);
  L2 = (IHZ*(1-pI)*GAMMA/(GAMMA+DELTA)) + (pL2*L2);
  // new infections
  IVZ_NEW = (E*(1-pE)*PHI/(DELTA+PHI));
  IHZ_NEW = (L1*(1-pL1)*(IOTA*kappa*psi)/(IOTA*kappa*psi+DELTA));
  // report rate for herpes zoster is time varying
  report_rate_hz +=  m2*exp(m1*t); //NEW EXPONENTIAL
  if(report_rate_hz > 1.0){report_rate_hz = 1.0;}
  if(report_rate_hz < 0.0){report_rate_hz = 0.0;}
}
